/*  
Program : cmeter.c, Colorimeter running on AtMega32 micro-controller
Listens on the RS232 port for commands fom the PC, and acts accordingly.
Author  : Ajith Kumar B.P, ( bpajith at gmail.com )
License : GNU GPL version 3
First Edit on 27-Mar-2013
*/

#include <avr/io.h>
#include <avr/pgmspace.h>

#define	GROUPSIZE	40	// Up to 40 commands in each group

// commands without any arguments (1 to 40)
#define GETVERSION	1	// Get the Eyes firmware version
#define READADC	    2

// Commands with One byte argument (41 to 80) 
#define SETLED		41
#define READCHAN	42

// Reply from ATmega8 to the PC
#define DONE		'D'	// Command executed successfully
#define	INVCMD		'C'	// Invalid Command
#define INVARG		'A'	// Invalid input data

const char version[] PROGMEM = "cm1.0";
typedef uint16_t  u16;
typedef uint8_t  u8;

#define	BUFSIZE		100		
//----------------------- Global variables -----------------------------
u8 	dbuf[BUFSIZE];		// status + data
u16 buf_index, tmp16;

void delay (uint16_t k)	 // roughly 2 usec per loop at 8 MHz system clock
{
  volatile uint16_t x = k;
  while (x)  --x;
}

int main()
{
u8 cmd, led;

// Initialize the RS232 communication link to the PC 38400, 8, 1, E
UCSRB = (1<<RXEN) | (1<<TXEN);  //UCSRA = (1 << U2X);
UBRRH = 0;
UBRRL = 12;						// At 8MHz (12 =>38400) (25 => 19200)
UCSRC = (1<<URSEL) | (1<<UPM1) | (1<<UCSZ1) | (1<<UCSZ0); // 8,1,E
ADCSRA = (1 << ADEN);

DDRA = 0;			// All ports as input
DDRC = 0;
DDRD = 0;
PORTA = 0;		  	// Disable all pullup 
PORTC = 0;
PORTD = 0;

buf_index = 0;
for(;;)
    {
    while ( !(UCSRA & (1<<RXC)) ) ;			// wait for receiver data
    dbuf[buf_index++] = UDR;		    	// Put the byte in the buffer.

    if(buf_index*GROUPSIZE > dbuf[0])	// Process after required no. of arguments
      {
	  cmd = dbuf[0];
	  dbuf[0] = 'D';			// Fill reply Assuming Success
	  buf_index = 1;			// Filling of return Data from second byte onwards
      switch(cmd)
		 {
	     case SETLED:			// Only for testing, to be remove later
      		led = dbuf[1];
			if( (led < 1) || (led > 20) ) 
				{
	      		dbuf[0] = INVARG;		// Invalid argument
	      		break;	
				}
			led = 21 - led;           // Wired in reverse order
			DDRA = 128;			// Disable current output
			DDRC = 0;
			DDRD = 0;
			PORTA = 128;
			if(led <= 6)  			// 1 to 6 on PORTD
				{
				DDRD = (1 << (led+1));
				}
			else if(led <= 14)  	// 7 to 14 on PORTC
				{
				DDRC = (1 << (led-7));
				}
			else   					// 15 to 20 on PORTA
				{
				DDRA = (1 << (led-15)) | 128;
				}
			PORTA &= ~128;    // enable current
		    break;

		#define ADCHAN	6
		case READCHAN:
			// Turn on the desired LED, Reads ADC, turns LED OFF, returns the voltage.
			// All LED
      		led = dbuf[1];
			if( (led < 1) || (led > 20) ) 
				{
	      		dbuf[0] = INVARG;		// Invalid argument
	      		break;	
				}
			led = 21 - led;           // Wired in reverse order
			DDRA = 128;			// Disable current output
			DDRC = 0;
			DDRD = 0;
			PORTA = 128;
			if(led <= 6)  			// 1 to 6 on PORTD
				{
				DDRD = (1 << (led+1));
				}
			else if(led <= 14)  	// 7 to 14 on PORTC
				{
				DDRC = (1 << (led-7));
				}
			else   					// 15 to 20 on PORTA
				{
				DDRA = (1 << (led-15)) | 128;
				}
			PORTA &= ~128;    							// enable current
			delay(2000);						// 4ms delay
	      	ADCSRA |= (1 <<ADIF);   					// clear old status
	      	ADMUX = (1 << 6) | ADCHAN;    				// with avcc
       	  	ADCSRA = (1 << ADEN) | (1 << ADSC) | 7;     // start, slow speed
      	  	while ( !(ADCSRA & (1<<ADIF)) ) ;           // wait for ADC conversion
          	dbuf[buf_index++] = ADCL;
          	dbuf[buf_index++] = ADCH;   
          	ADCSRA &= ~(1<<ADIF);
			DDRA = 128;									// Disable current output
			PORTA = 128;
		    break;

		case READADC:
	      	ADCSRA |= (1 <<ADIF);   					// clear old status
	      	ADMUX = (1 << 6) | ADCHAN;   				// with avcc
       	  	ADCSRA = (1 << ADEN) | (1 << ADSC) | 7;     // start, slow speed
      	  	while ( !(ADCSRA & (1<<ADIF)) ) ;           // wait for ADC conversion
          	dbuf[buf_index++] = ADCL;
          	dbuf[buf_index++] = ADCH;   
          	ADCSRA &= ~(1<<ADIF);
		    break;

	     
		 case GETVERSION:
      		memcpy_P(&dbuf[1], version,5);
      		buf_index += 5;
      		break;

    	 default:
      		dbuf[0] = INVCMD;		// Invalid Command
      		break;	
         }

   	  while( !(UCSRA & (1 <<UDRE) ) );
      UDR = dbuf[0];				// Send the response to PC in any case  					
	  if(dbuf[0] == 'D')			// If no error, send the data too, no handshake.
	 	 for(tmp16=1; tmp16 < buf_index; ++tmp16)	
      		{
    		while( !(UCSRA & (1 <<UDRE) ) );
    		UDR = dbuf[tmp16];
			}
      buf_index = 0;
      }
    }
return 0;
}

